@@ -29,35 +29,24 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: canon.c,v 1.1.1.1 2003/08/07 16:46:05 ccpro Exp $
+ * $Id: canon.c,v 1.38 2003/08/08 20:40:55 ejohnst Exp $
  */
 
 /*
  * Exif tag definitions for Canon maker notes.
  * Developed from http://www.burren.cx/david/canon.html.
  * EOS 1D and 1Ds contributions from Stan Jirman <stanj@mac.com>.
+ * EOS 10D contributions from Jason Montojo <jason.montojo@rogers.com>.
  *
  */
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <errno.h>
 
-#include "exif.h"
-#include "exifint.h"
 #include "makers.h"
 
 
-/* Custom function field description lookup table. */
-
-struct ccstm {
-	int32_t val;
-	const char *descr;
-	struct descrip *table;
-};
-
-
 /* Macro mode. */
 
 static struct descrip canon_macro[] = {
@@ -72,6 +61,7 @@ static struct descrip canon_macro[] = {
 static struct descrip canon_focustype[] = {
 	{ 0,	"Manual" },
 	{ 1,	"Auto" },
+	{ 2,	"Auto" },
 	{ 3,	"Close-Up (Macro Mode)" },
 	{ 7,	"Infinity Mode" },
 	{ 8,	"Locked (Pan Mode)" },
@@ -276,7 +266,7 @@ static struct exiftag canon_tags[] = {
 	  "Custom Function", NULL },
 	{ 0x00a0, TIFF_SHORT, 0,  ED_UNK, "CanonA0Tag",
 	  "Canon TagA0 Offset", NULL },
-	{ 0xffff, TIFF_UNKN,  0,  ED_UNK, "Unknown",
+	{ 0xffff, TIFF_UNKN,  0,  ED_UNK, "CanonUnknown",
 	  "Canon Unknown", NULL },
 };
 
@@ -334,7 +324,7 @@ static struct exiftag canon_tags01[] = {
 	  "Zoomed Resolution", NULL },
 	{ 37, TIFF_SHORT, 0, ED_VRB, "CanonBZoomRes",
 	  "Base Zoom Resolution", NULL },
-	{ 0xffff, TIFF_SHORT, 0, ED_UNK, "CanonUnknown",
+	{ 0xffff, TIFF_SHORT, 0, ED_UNK, "Canon01Unknown",
 	  "Canon Tag1 Unknown", NULL },
 };
 
@@ -354,7 +344,7 @@ static struct exiftag canon_tags04[] = {
 	  "Flash Bias", canon_fbias },
 	{ 19, TIFF_SHORT, 0, ED_UNK, "CanonSubjDst",
 	  "Subject Distance", NULL },
-	{ 0xffff, TIFF_SHORT, 0, ED_UNK, "CanonUnknown",
+	{ 0xffff, TIFF_SHORT, 0, ED_UNK, "Canon04Unknown",
 	  "Canon Tag4 Unknown", NULL },
 };
 
@@ -364,12 +354,22 @@ static struct exiftag canon_tags04[] = {
 static struct exiftag canon_tagsA0[] = {
 	{ 9,  TIFF_SHORT, 0, ED_IMG, "CanonColorTemp",
 	  "Color Temperature", NULL },
-	{ 0xffff, TIFF_SHORT, 0, ED_UNK, "CanonUnknown",
+	{ 10, TIFF_SHORT, 0, ED_IMG, "CanonColorMatrix",
+	  "Color Matrix", NULL },
+	{ 0xffff, TIFF_SHORT, 0, ED_UNK, "CanonA0Unknown",
 	  "Canon TagA0 Unknown", NULL },
 };
 
 
-/* Value descriptions for D30, D60 custom functions. */
+/* Placeholder for unknown fields. */
+
+static struct exiftag canon_tagsunk[] = {
+	{ 0xffff, TIFF_SHORT, 0, ED_UNK, "CanonUnknown",
+	  "Canon Unknown", NULL },
+};
+
+
+/* Value descriptions for custom functions. */
 
 static struct descrip ccstm_offon[] = {
 	{ 0,	"Off" },
@@ -457,6 +457,12 @@ static struct descrip ccstm_yesno[] = {
 	{ -1,	"Unknown" },
 };
 
+static struct descrip ccstm_noyes[] = {
+	{ 0,	"No" },
+	{ 1,	"Yes" },
+	{ -1,	"Unknown" },
+};
+
 static struct descrip ccstm_onoff[] = {
 	{ 0,	"On" },
 	{ 1,	"Off" },
@@ -564,60 +570,196 @@ static struct descrip ccstm_fscr[] = {
 	{ -1,	"Unknown" },
 };
 
-static struct descrip ccstm_finder[] = {
-	{ 0,	"No Viewfinder Display" },
-	{ 1,	"Finder Display On" },
+static struct descrip ccstm_10dsetbut[] = {
+	{ 0,	"Not Assigned" },
+	{ 1,	"Change Quality" },
+	{ 2,	"Change Parameters" },
+	{ 3,	"Menu Display" },
+	{ 4,	"Image Replay" },
+	{ -1,	"Unknown" },
+};
+
+static struct descrip ccstm_10dshutter[] = {
+	{ 0,	"AF/AE Lock" },
+	{ 1,	"AE Lock/AF" },
+	{ 2,	"AF/AF Lock, No AE Lock" },
+	{ 3,	"AE/AF, No AE Lock" },
+	{ -1,	"Unknown" },
+};
+
+static struct descrip ccstm_assistflash[] = {
+	{  0,	"Emits/Fires" },
+	{  1,	"Does Not Emit/Fires" },
+	{  2,	"Only Ext. Flash Emits/Fires" },
+	{  3,	"Emits/Does Not Fire" },
+	{ -1,	"Unknown" },
+};
+
+static struct descrip ccstm_afptreg[] = {
+	{ 0,	"Center" },
+	{ 1,	"Bottom" },
+	{ 2,	"Right" },
+	{ 3,	"Extreme Right" },
+	{ 4,	"Automatic" },
+	{ 5,	"Extreme Left" },
+	{ 6,	"Left" },
+	{ 7,	"Top" },
+	{ -1,	"Unknown" },
+};
+
+static struct descrip ccstm_rawjpeg[] = {
+	{ 0,	"RAW+Small/Normal" },
+	{ 1,	"RAW+Small/Fine" },
+	{ 2,	"RAW+Medium/Normal" },
+	{ 3,	"RAW+Medium/Fine" },
+	{ 4,	"RAW+Large/Normal" },
+	{ 5,	"RAW+Large/Fine" },
+	{ -1,	"Unknown" },
+};
+
+static struct descrip ccstm_10dmenubut[] = {
+	{ 0,	"Previous (Volatile)" },
+	{ 1,	"Previous" },
+	{ 2,	"Top" },
+	{ -1,	"Unknown" },
+};
+
+static struct descrip ccstm_assistbut[] = {
+	{ 0,	"Normal" },
+	{ 1,	"Select Home Position" },
+	{ 2,	"Select HP (while pressing)" },
+	{ 3,	"Av+/- (AF point by QCD)" },
+	{ 4,	"FE lock" },
 	{ -1,	"Unknown" },
 };
 
 
 /* D30/D60 custom functions. */
 
-static struct ccstm canon_d30custom[] = {
-	{ 1,	"Long exposure noise reduction", ccstm_offon },
-	{ 2,	"Shutter/AE lock buttons", ccstm_shutter },
-	{ 3,	"Mirror lockup", ccstm_disen },
-	{ 4,	"Tv/Av and exposure level", ccstm_explvl },
-	{ 5,	"AF-assist light", ccstm_autooff },
-	{ 6,	"Av mode shutter speed", ccstm_shutspd },
-	{ 7,	"AEB sequence/auto cancellation", ccstm_aebseq },
-	{ 8,	"Shutter curtain sync", ccstm_shutsync },
-	{ 9,	"Lens AF stop button", ccstm_lensaf },
-	{ 10,	"Fill flash auto reduction", ccstm_endis },
-	{ 11,	"Menu button return position", ccstm_menubut },
-	{ 12,	"Shooting Set button function", ccstm_setbut },
-	{ 13,	"Sensor cleaning", ccstm_disen },
-	{ 14,	"Superimposed display", ccstm_onoff },
-	{ 15,	"Shutter release w/o CF card", ccstm_yesno },
-	{ -1,	"Unknown function", NULL },
+static struct exiftag canon_d30custom[] = {
+	{ 1,  TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "Long exposure noise reduction", ccstm_offon },
+	{ 2,  TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "Shutter/AE lock buttons", ccstm_shutter },
+	{ 3,  TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "Mirror lockup", ccstm_disen },
+	{ 4,  TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "Tv/Av and exposure level", ccstm_explvl },
+	{ 5,  TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "AF-assist light", ccstm_autooff },
+	{ 6,  TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "Av mode shutter speed", ccstm_shutspd },
+	{ 7,  TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "AEB sequence/auto cancellation", ccstm_aebseq },
+	{ 8,  TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "Shutter curtain sync", ccstm_shutsync },
+	{ 9,  TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "Lens AF stop button", ccstm_lensaf },
+	{ 10, TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "Fill flash auto reduction", ccstm_endis },
+	{ 11, TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "Menu button return position", ccstm_menubut },
+	{ 12, TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "Shooting Set button function", ccstm_setbut },
+	{ 13, TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "Sensor cleaning", ccstm_disen },
+	{ 14, TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "Superimposed display", ccstm_onoff },
+	{ 15, TIFF_SHORT, 0, ED_VRB, "D30Custom",
+	  "Shutter release w/o CF card", ccstm_yesno },
+	{ 0xffff, TIFF_SHORT, 0, ED_UNK, "D30CustomUnknown",
+	  "Canon D30/D60 Custom Unknown", NULL },
 };
 
 
 /* EOS-1D/1Ds custom functions. */
 
-static struct ccstm canon_1dcustom[] = {
-	{ 0,	"Focusing screen", ccstm_fscr },
-	{ 1,	"Finder display during exposure", ccstm_finder },
-	{ 2,	"Shutter release w/o CF card", ccstm_yesno },
-	{ 3,	"ISO speed expansion", ccstm_yesno },
-	{ 4,	"Shutter button/AEL button", ccstm_shutterael },
-	{ 5,	"Manual Tv/Av for M", ccstm_tvavform },
-	{ 6,	"Exposure level increments", ccstm_explvlinc },
-	{ 7,	"USM lens electronic MF", ccstm_usmmf },
-	{ 8,	"Top/back LCD panels", ccstm_lcdpanels },
-	{ 9,	"AEB sequence/auto cancellation", ccstm_aebseq },
-	{ 10,	"AF point illumination", ccstm_afill },
-	{ 11,	"AF point selection", ccstm_afsel },
-	{ 12,	"Mirror lockup", ccstm_disen },
-	{ 13,	"# AF points/spot metering", ccstm_afspot },
-	{ 14,	"Fill flash auto reduction", ccstm_endis },
-	{ 15,	"Shutter curtain sync", ccstm_shutsync },
-	{ 16,	"Safety shift in Av or Tv", ccstm_endis },
-	{ 17,	"AF point activation area", ccstm_afact },
-	{ 18,	"Switch to registered AF point", ccstm_regaf },
-	{ 19,	"Lens AF stop button", ccstm_lensaf1 },
-	{ 20,	"AI servo tracking sensitivity", ccstm_aisens },
-	{ -1,	"Unknown function", NULL },
+static struct exiftag canon_1dcustom[] = {
+	{ 0,  TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "Focusing screen", ccstm_fscr },
+	{ 1,  TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "Finder display during exposure", ccstm_offon },
+	{ 2,  TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "Shutter release w/o CF card", ccstm_yesno },
+	{ 3,  TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "ISO speed expansion", ccstm_noyes },
+	{ 4,  TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "Shutter button/AEL button", ccstm_shutterael },
+	{ 5,  TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "Manual Tv/Av for M", ccstm_tvavform },
+	{ 6,  TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "Exposure level increments", ccstm_explvlinc },
+	{ 7,  TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "USM lens electronic MF", ccstm_usmmf },
+	{ 8,  TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "Top/back LCD panels", ccstm_lcdpanels },
+	{ 9,  TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "AEB sequence/auto cancellation", ccstm_aebseq },
+	{ 10, TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "AF point illumination", ccstm_afill },
+	{ 11, TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "AF point selection", ccstm_afsel },
+	{ 12, TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "Mirror lockup", ccstm_disen },
+	{ 13, TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "# AF points/spot metering", ccstm_afspot },
+	{ 14, TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "Fill flash auto reduction", ccstm_endis },
+	{ 15, TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "Shutter curtain sync", ccstm_shutsync },
+	{ 16, TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "Safety shift in Av or Tv", ccstm_disen },
+	{ 17, TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "AF point activation area", ccstm_afact },
+	{ 18, TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "Switch to registered AF point", ccstm_regaf },
+	{ 19, TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "Lens AF stop button", ccstm_lensaf1 },
+	{ 20, TIFF_SHORT, 0, ED_VRB, "1DCustom",
+	  "AI servo tracking sensitivity", ccstm_aisens },
+	{ 0xffff, TIFF_SHORT, 0, ED_UNK, "1DCustomUnknown",
+	  "Canon 1D/1Ds Custom Unknown", NULL },
+};
+
+/* 10D custom functions. */
+
+static struct exiftag canon_10dcustom[] = {
+	{ 1,  TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "SET button function when shooting", ccstm_10dsetbut },
+	{ 2,  TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "Shutter release w/o CF card", ccstm_yesno },
+	{ 3,  TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "Flash sync speed in Av mode", ccstm_shutspd },
+	{ 4,  TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "Shutter button/AE lock button", ccstm_10dshutter },
+	{ 5,  TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "AF-assist beam/Flash firing", ccstm_assistflash },
+	{ 6,  TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "Exposure level increments", ccstm_explvl },
+	{ 7,  TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "AF point registration", ccstm_afptreg },
+	{ 8,  TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "RAW+JPEG recording", ccstm_rawjpeg },
+	{ 9,  TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "AEB sequence/auto cancellation", ccstm_aebseq },
+	{ 10, TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "Superimposed display", ccstm_onoff },
+	{ 11, TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "Menu button display position", ccstm_10dmenubut },
+	{ 12, TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "Mirror lockup", ccstm_disen },
+	{ 13, TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "Assist button function", ccstm_assistbut },
+	{ 14, TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "Fill flash auto reduction", ccstm_endis },
+	{ 15, TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "Shutter curtain sync", ccstm_shutsync },
+	{ 16, TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "Safety shift in Av or Tv", ccstm_disen },
+	{ 17, TIFF_SHORT, 0, ED_VRB, "10DCustom",
+	  "Lens AF stop button", ccstm_lensaf },
+	{ 0xffff, TIFF_SHORT, 0, ED_UNK, "10DCustomUnknown",
+	  "Canon 10D Custom Unknown", NULL },
 };
 
 
@@ -625,18 +767,16 @@ static struct ccstm canon_1dcustom[] = {
  * Process maker note tag 0x0001 values.
  */
 static int
-canon_prop01(struct exifprop *aprop, struct exifprop *prop, char *off,
-    enum order o)
+canon_prop01(struct exifprop *aprop, struct exifprop *prop,
+    unsigned char *off, enum byteorder o)
 {
 	u_int16_t v = (u_int16_t)aprop->value;
 
-	switch (aprop->subtag) {
+	switch (aprop->tag) {
 	case 2:
 		aprop->lvl = v ? ED_IMG : ED_VRB;
-		if (!(aprop->str = (char *)malloc(32)))
-			exifdie((const char *)strerror(errno));
+		exifstralloc(&aprop->str, 32);
 		snprintf(aprop->str, 31, "%d sec", v / 10);
-		aprop->str[31] = '\0';
 		break;
 	case 5:
 		/* Change "Single" to "Timed" if #2 > 0. */
@@ -653,12 +793,10 @@ canon_prop01(struct exifprop *aprop, struct exifprop *prop, char *off,
 		 */
 
 		if (v == 3 && prop->count >= 37) {
-			if (!(aprop->str = (char *)malloc(32)))
-				exifdie((const char *)strerror(errno));
+			exifstralloc(&aprop->str, 32);
 			snprintf(aprop->str, 31, "x%.1f", 2 *
 			    (float)exif2byte(off + 37 * 2, o) /
 			    (float)exif2byte(off + 36 * 2, o));
-			aprop->str[31] = '\0';
 		} else
 			aprop->str = finddescr(canon_dzoom, v);
 		break;
@@ -690,11 +828,11 @@ canon_prop01(struct exifprop *aprop, struct exifprop *prop, char *off,
  * Process maker note tag 0x0004 values.
  */
 static int
-canon_prop04(struct exifprop *aprop, struct exifprop *prop, char *off,
-    enum order o)
+canon_prop04(struct exifprop *aprop, struct exifprop *prop,
+    unsigned char *off, enum byteorder o)
 {
 
-	switch (aprop->subtag) {
+	switch (aprop->tag) {
 	case 7:
 		aprop->override = EXIF_T_WHITEBAL;
 		break;
@@ -713,16 +851,14 @@ canon_prop04(struct exifprop *aprop, struct exifprop *prop, char *off,
  * Process maker note tag 0x00a0 values.
  */
 static int
-canon_propA0(struct exifprop *aprop, struct exifprop *prop, char *off,
-    enum order o)
+canon_propA0(struct exifprop *aprop, struct exifprop *prop,
+    unsigned char *off, enum byteorder o)
 {
 
-	switch (aprop->subtag) {
+	switch (aprop->tag) {
 	case 9:
-		if (!(aprop->str = (char *)malloc(32)))
-			exifdie((const char *)strerror(errno));
+		exifstralloc(&aprop->str, 32);
 		snprintf(aprop->str, 31, "%d K", aprop->value);
-		aprop->str[31] = '\0';
 		break;
 	default:
 		return (FALSE);
@@ -743,21 +879,26 @@ canon_subval(struct exifprop *prop, struct exiftags *t,
 	int i, j;
 	u_int16_t v;
 	struct exifprop *aprop;
-	char *off = t->btiff + prop->value;
+	unsigned char *off = t->md.btiff + prop->value;
 
-	/* Check size of tag (first value). */
+	/* Check size of tag (first value) if we're not debugging. */
 
-	if (exif2byte(off, t->tifforder) != 2 * prop->count) {
+	if (valfun && exif2byte(off, t->md.order) != 2 * prop->count) {
 		exifwarn("Canon maker tag appears corrupt");
 		return (FALSE);
 	}
 
+	if (debug)
+		printf("Processing %s (0x%04X) directory, %d entries\n",
+		    prop->name, prop->tag, prop->count);
+
 	for (i = 0; i < (int)prop->count; i++) {
-		v = exif2byte(off + i * 2, t->tifforder);
+		v = exif2byte(off + i * 2, t->md.order);
 
 		aprop = childprop(prop);
 		aprop->value = (u_int32_t)v;
-		aprop->subtag = i;
+		aprop->tag = i;
+		aprop->tagset = subtags;
 
 		/* Lookup property name and description. */
 
@@ -776,13 +917,13 @@ canon_subval(struct exifprop *prop, struct exiftags *t,
 		if (valfun && !valfun(aprop, prop, off, t)) {
 			if (aprop->lvl != ED_UNK)
 				continue;
-
-			if (!(aprop->str = (char *)malloc(32)))
-				exifdie((const char *)strerror(errno));
+			exifstralloc(&aprop->str, 32);
 			snprintf(aprop->str, 31, "num %02d, val 0x%04X", i, v);
-			aprop->str[31] = '\0';
 		}
 	}
+
+	if (debug)
+		printf("\n");
 	return (TRUE);
 }
 
@@ -791,10 +932,10 @@ canon_subval(struct exifprop *prop, struct exiftags *t,
  * Process custom function tag values.
  */
 static void
-canon_custom(struct exifprop *prop, char *off, enum order o,
-    struct ccstm *table)
+canon_custom(struct exifprop *prop, unsigned char *off, enum byteorder o,
+    struct exiftag *table)
 {
-	int i, j;
+	int i, j = -1;
 	const char *cn;
 	char *cv = NULL;
 	u_int16_t v;
@@ -813,36 +954,38 @@ canon_custom(struct exifprop *prop, char *off, enum order o,
 		return;
 	}
 
+	if (debug)
+		printf("Processing %s directory, %d entries\n", prop->name,
+		    prop->count);
+
 	for (i = 1; i < (int)prop->count; i++) {
 		v = exif2byte(off + i * 2, o);
 
 		aprop = childprop(prop);
-		aprop->value = (u_int32_t)v;
-		aprop->subtag = i;
-		aprop->name = prop->name;
+		aprop->value = v & 0xff;
+		aprop->tag = v >> 8 & 0xff;
+		aprop->tagset = table;
+
+		/*
+		 * Lookup function name and value.  First byte is function
+		 * number; second is function value.
+		 */
+
+		for (j = 0; table[j].tag != EXIF_T_UNKNOWN &&
+		    table[j].tag != (v >> 8 & 0xff); j++);
+		aprop->name = table[j].name;
 		aprop->descr = prop->descr;
-		aprop->lvl = ED_VRB;
+		aprop->lvl = table[j].lvl;
+		if (table[j].table)
+			cv = finddescr(table[j].table,
+			    (u_int16_t)(v & 0xff));
+		cn = table[j].descr;
+
 
 		dumpprop(aprop, NULL);
 
-		/*
-		 * If we have a table, lookup function name and value.
-		 * First byte is function number; second is function value.
-		 */
-
-		if (table) {
-			for (j = 0; table[j].val != -1 &&
-			    table[j].val != (v >> 8 & 0xff); j++);
-			if (table[j].table)
-				cv = finddescr(table[j].table,
-				    (u_int16_t)(v & 0xff));
-			cn = table[j].descr;
-		} else
-			cn = "Unknown";
-
-		if (!(aprop->str = (char *)malloc(4 + strlen(cn) +
-		    (cv ? strlen(cv) : 10))))
-			exifdie((const char *)strerror(errno));
+		exifstralloc(&aprop->str, 4 + strlen(cn) +
+		    (cv ? strlen(cv) : 10));
 
 		if (cv && j != -1) {
 			snprintf(aprop->str, 4 + strlen(cn) + strlen(cv),
@@ -856,6 +999,9 @@ canon_custom(struct exifprop *prop, char *off, enum order o,
 			aprop->lvl = ED_UNK;
 		}
 	}
+
+	if (debug)
+		printf("\n");
 }
 
 
@@ -865,37 +1011,10 @@ canon_custom(struct exifprop *prop, char *off, enum order o,
 void
 canon_prop(struct exifprop *prop, struct exiftags *t)
 {
-	unsigned int i;
-	char *offset;
-	u_int16_t v, flmin = 0, flmax = 0, flunit = 0;
+	unsigned char *offset;
+	u_int16_t flmin = 0, flmax = 0, flunit = 0;
 	struct exifprop *tmpprop;
 
-	/*
-	 * Don't process child properties we've created while looking at
-	 * other maker note tags.
-	 */
-
-	if (prop->subtag > -2)
-		return;
-
-	/* Lookup the field name (if known). */
-
-	for (i = 0; canon_tags[i].tag < EXIF_T_UNKNOWN &&
-	    canon_tags[i].tag != prop->tag; i++);
-	prop->name = canon_tags[i].name;
-	prop->descr = canon_tags[i].descr;
-	prop->lvl = canon_tags[i].lvl;
-
-	if (debug) {
-		static int once = 0;
-
-		if (!once) {
-			printf("Processing Canon Maker Note\n");
-			once = 1;
-		}
-		dumpprop(prop, NULL);
-	}
-
 	switch (prop->tag) {
 
 	/* Various image data. */
@@ -911,18 +1030,17 @@ canon_prop(struct exifprop *prop, struct exiftags *t)
 		 */
 
 		if (prop->count >= 25) {
-			offset = t->btiff + prop->value;
-			flmax = exif2byte(offset + 23 * 2, t->tifforder);
-			flmin = exif2byte(offset + 24 * 2, t->tifforder);
-			flunit = exif2byte(offset + 25 * 2, t->tifforder);
+			offset = t->md.btiff + prop->value;
+			flmax = exif2byte(offset + 23 * 2, t->md.order);
+			flmin = exif2byte(offset + 24 * 2, t->md.order);
+			flunit = exif2byte(offset + 25 * 2, t->md.order);
 		}
 
 		if (flunit && (flmin || flmax)) {
 			tmpprop = childprop(prop);
 			tmpprop->name = "CanonLensSz";
 			tmpprop->descr = "Lens Size";
-			if (!(tmpprop->str = (char *)malloc(32)))
-				exifdie((const char *)strerror(errno));
+			exifstralloc(&tmpprop->str, 32);
 
 			if (flmin == flmax) {
 				snprintf(tmpprop->str, 31, "%.2f mm",
@@ -947,9 +1065,9 @@ canon_prop(struct exifprop *prop, struct exiftags *t)
 
 		/* Color temp is bad if white balance isn't manual. */
 
-		if ((tmpprop = findsprop(t->props, 0x0004, 7)))
+		if ((tmpprop = findprop(t->props, canon_tags04, 7)))
 			if (tmpprop->value != 9) {
-				if ((tmpprop = findsprop(prop, 0x00a0, 9)))
+				if ((tmpprop = findprop(prop, canon_tagsA0, 9)))
 					tmpprop->lvl = ED_BAD;
 		}
 		break;
@@ -957,30 +1075,46 @@ canon_prop(struct exifprop *prop, struct exiftags *t)
 	/* Image number. */
 
 	case 0x0008:
-		if (!(prop->str = (char *)malloc(32)))
-			exifdie((const char *)strerror(errno));
+		exifstralloc(&prop->str, 32);
 		snprintf(prop->str, 31, "%03d-%04d", prop->value / 10000,
 		    prop->value % 10000);
-		prop->str[31] = '\0';
 		break;
 
 	/* Serial number. */
 
 	case 0x000c:
-		if (!(prop->str = (char *)malloc(11)))
-			exifdie((const char *)strerror(errno));
+		exifstralloc(&prop->str, 11);
 		snprintf(prop->str, 11, "%010d", prop->value);
 		break;
 
 	/* Custom functions. */
 
 	case 0x000f:
-		canon_custom(prop, t->btiff + prop->value, t->tifforder,
-		    canon_d30custom);
+		/*
+		 * Canon annoyingly reuses this tag value for different sets
+		 * of custom functions (e.g., D30/60, 10D).  Therefore, we
+		 * won't try to interpret them unless we know for sure that
+		 * the camera model is supported.
+		 */
+
+		if (!t->model) {
+			exifwarn("Canon model unset; please report to author");
+			break;
+		}
+
+		if (strstr(t->model, "10D"))
+			canon_custom(prop, t->md.btiff + prop->value,
+			    t->md.order, canon_10dcustom);
+		else if (strstr(t->model, "D30") || strstr(t->model, "D60"))
+			canon_custom(prop, t->md.btiff + prop->value,
+			    t->md.order, canon_d30custom);
+		else
+			exifwarn2("Custom function unsupported for %s; please "
+			    "report to author", t->model);
 		break;
 
 	case 0x0090:
-		canon_custom(prop, t->btiff + prop->value, t->tifforder,
+		canon_custom(prop, t->md.btiff + prop->value, t->md.order,
 		    canon_1dcustom);
 		break;
 
@@ -988,12 +1122,18 @@ canon_prop(struct exifprop *prop, struct exiftags *t)
 
 	default:
 		if (prop->type == TIFF_SHORT && prop->count > 1 && debug)
-			for (i = 0; i < prop->count; i++) {
-				v = exif2byte(t->btiff + prop->value +
-				    (i * 2), t->tifforder);
-				printf("     Unknown (%d): %d, 0x%04X\n",
-				    i, v, v);
-			}
+			canon_subval(prop, t, canon_tagsunk, NULL);
 		break;
 	}
 }
+
+
+/*
+ * Try to read Canon maker note IFDs.
+ */
+struct ifd *
+canon_ifd(u_int32_t offset, struct tiffmeta *md)
+{
+
+	return(readifds(offset, canon_tags, md));
+}
